#include "tilemap.h"
#include "camera.h"
#include "program.h"
#include <stdio.h>
#include <libxml/tree.h>
#include <SDL2/SDL_image.h>

#define XML(__str) (const xmlChar*)__str

static inline
void tilemap_get_size_from_xml(xmlNodePtr node, int* o_width, int* o_height) {
    xmlChar* prop = xmlGetProp(node, XML("width"));
    if(prop == NULL) { printf("Nu-uh\n"); return; }
    *o_width = atoi((char*)prop);

    prop = xmlGetProp(node, XML("height"));
    if(prop == NULL) { printf("Nu-uh\n"); return; }
    *o_height = atoi((char*)prop);
}

static inline
void tilemap_get_tiles_from_xml(xmlDoc* doc, xmlNodePtr node, int* out, int out_len) {
    node = node->children;
    while(xmlStrcmp(node->name, XML("data")) != 0 && node != NULL) {
        node = node->next;
    }

    if(node == NULL) {
        memset(out, 0x0, out_len * sizeof(int));
        return;
    }

    char buffer[5]; buffer[4] = '\0';
    char* buffer_writer = buffer;
    int* out_writer = out;
    xmlChar* str = xmlNodeGetContent(node);
    xmlChar* reader = str;

    while(*reader != '\0') {
       while(isspace(*reader)) 
            reader++;
        if(*reader != ',' && *reader != '\0') {
            *buffer_writer = (char)*reader;
            ++buffer_writer;
        } else {
            *buffer_writer = '\0';
            buffer_writer = buffer;
            *out_writer = atoi(buffer) - 1;
            ++out_writer;
        }
        ++reader;
    }
}

struct Tilemap tilemap_load(const char* tilemap_file) {
    struct Tilemap self = {
        .dimensions = {1,1},
        .tiles = NULL
    };

    xmlDocPtr doc = xmlParseFile(tilemap_file);
    if(doc == NULL) {
        printf("Failed to load tilemap file '%s'\n", tilemap_file);
        self.dimensions.x = self.dimensions.y = 0;
        return self;
    }

    xmlNodePtr reader = xmlDocGetRootElement(doc);
    if(reader == NULL) {
        printf("Failed to find map node in tilemap '%s'\n", tilemap_file);
        self.dimensions.x = self.dimensions.y = 0;
        xmlFreeDoc(doc);
        return self;
    }

    reader = reader->children;

    do {
        if(xmlStrcmp(reader->name, XML("layer")) == 0) {
            tilemap_get_size_from_xml(reader, &self.dimensions.x, &self.dimensions.y);
            printf("sizes: %d %d\n", self.dimensions.x, self.dimensions.y);
            self.tiles = malloc(self.dimensions.x * self.dimensions.y * sizeof(int));
            printf("reading layer data\n");
            tilemap_get_tiles_from_xml(doc, reader, self.tiles, self.dimensions.x * self.dimensions.y);
        } else if(xmlStrcmp(reader->name, XML("objectgroup"))) {
        }
        reader = reader->next;
    } while(reader != NULL);

    xmlFreeDoc(doc);
    
    return self;
}

static inline
int is_path_char(int n) {
    return isalnum(n) || n == '/';
}

struct Tileset tileset_load(const char* filename) {
    char* texture_path = NULL;
    char* writer = NULL;
    char num_buffer[6];
    num_buffer[5] = '\0';
    int n = 0;
    struct Tileset self = {
        .texture = NULL,
        .tile_size = {0,0}
    };
    size_t len = 1;

    FILE* fs = fopen(filename, "r");

    do {
        n = fgetc(fs);
        if(is_path_char(n)) {
            ++len;
        }
    } while(n != ',');

    texture_path = malloc(len * sizeof(char));

    rewind(fs);
    writer = texture_path;

    do {
        n = fgetc(fs);
        if(is_path_char(n)) {
            *writer = n;
            ++writer;
        } else if(n == ',') {
            *writer = '\0';
        }
    } while(n != ',');

    self.texture = IMG_LoadTexture(g_renderer, texture_path);
    SDL_QueryTexture(self.texture, NULL, NULL, &self.texture_resolution.x, &self.texture_resolution.y);
    self.shear = floor((float)self.texture_resolution.x / self.tile_size.x);

    writer = num_buffer;

    do {
        n = fgetc(fs);
        if(isdigit(n) || n == '-') {
            *writer= n;
            ++writer;
        } else if(n == ',' || feof(fs) || n == '\n') {
            *writer = '\0';
            writer = num_buffer;
            if(self.tile_size.x == 0) {
                self.tile_size.x = atoi(num_buffer);
            } else {
                self.tile_size.y = atoi(num_buffer);
            }
        }
    } while(!feof(fs));

    fclose(fs);

    return self;
}

SDL_Rect tileset_index_to_rect(struct Tileset* self, size_t index) {
    SDL_Rect rect = {0,0, self->tile_size.x, self->tile_size.y};
    rect.x = (index % self->shear) * rect.w;
    rect.y = (index / self->shear) * rect.h;
    return rect;
}

void tilemap_render(struct Tilemap* self) {
    size_t num_tiles = self->dimensions.x * self->dimensions.y;

    SDL_Rect source_rect = {0, 0, self->tileset.tile_size.x, self->tileset.tile_size.y};
    SDL_FRect dest_rect = { .w = 1, .h = 1 };

    for(int i = 0; i < num_tiles; ++i) {
        source_rect = tileset_index_to_rect(&self->tileset, self->tiles[i]);

        dest_rect.x = (i % self->dimensions.x) * dest_rect.w;
        dest_rect.y = (float)floor((float)i / self->dimensions.y) * dest_rect.h;

        SDL_RenderCopyF(g_renderer, self->tileset.texture, &source_rect, &dest_rect);
    }
}